/**
 * @file   ModbusRtu.h
 * @version     1.21
 * @date        2016.02.21
 * @author  Samuel Marco i Armengol
 * @contact     sammarcoarmengol@gmail.com
 * @contribution Helium6072
 *
 * @description
 *  Arduino library for communicating with Modbus devices
 *  over RS232/USB/485 via RTU protocol.
 *
 *  Further information:
 *  http://modbus.org/
 *  http://modbus.org/docs/Modbus_over_serial_line_V1_02.pdf
 *
 * @license
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; version
 *  2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * @defgroup setup Modbus Object Instantiation/Initialization
 * @defgroup loop Modbus Object Management
 * @defgroup buffer Modbus Buffer Management
 * @defgroup discrete Modbus Function Codes for Discrete Coils/Inputs
 * @defgroup register Modbus Function Codes for Holding/Input Registers
 *
 */

/******************************************************************************
     * Based on the Modbus Master-Slave library for Arduino. 
     * Author project library - smarmengol.
     * Version on commit 8321f2dc15535d8cecd4a08331aaf112f171c31c on 30 Jul 2016.
     * Modbus Master-Slave library for Arduino - https://github.com/smarmengol/Modbus-Master-Slave-for-Arduino. 
     * GNU Lesser General Public License v2.1
******************************************************************************/

/******************************************************************************
* File: ModbusRtuSlave.cpp
* Description: Библиотека работы с Modbus RTU в режиме Slave.
* Created: 22 ноября 2016
* Last Modifiede: 06 декабря 2016
* Copyright: GNU General Public License
******************************************************************************/

/*
//TODO: Добавить комментарии.
//TODO: Сделать буфер для отложенной записи данных и реализацию для отложенной записи в EEPROM.
*/

//*******************************  Библиотеки  ******************************//
#include "ModbusRtuSlave.h"

#ifdef USE_AVR_UART
	#include "UART.h"
#endif

//#include "UARTDebug.h"
//*******************************  /Библиотеки  *****************************//

/* Table of CRC values for high–order byte */
static const uint8_t auchCRCHi[] PROGMEM = {
	0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81,
	0x40, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0,
	0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01,
	0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41,
	0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81,
	0x40, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0,
	0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01,
	0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40,
	0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81,
	0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0,
	0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01,
	0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41,
	0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81,
	0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0,
	0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01,
	0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41,
	0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81,
	0x40
} ; 

/* Table of CRC values for low–order byte */
static const int8_t auchCRCLo[] PROGMEM = {
	0x00, 0xC0, 0xC1, 0x01, 0xC3, 0x03, 0x02, 0xC2, 0xC6, 0x06, 0x07, 0xC7, 0x05, 0xC5, 0xC4,
	0x04, 0xCC, 0x0C, 0x0D, 0xCD, 0x0F, 0xCF, 0xCE, 0x0E, 0x0A, 0xCA, 0xCB, 0x0B, 0xC9, 0x09,
	0x08, 0xC8, 0xD8, 0x18, 0x19, 0xD9, 0x1B, 0xDB, 0xDA, 0x1A, 0x1E, 0xDE, 0xDF, 0x1F, 0xDD,
	0x1D, 0x1C, 0xDC, 0x14, 0xD4, 0xD5, 0x15, 0xD7, 0x17, 0x16, 0xD6, 0xD2, 0x12, 0x13, 0xD3,
	0x11, 0xD1, 0xD0, 0x10, 0xF0, 0x30, 0x31, 0xF1, 0x33, 0xF3, 0xF2, 0x32, 0x36, 0xF6, 0xF7,
	0x37, 0xF5, 0x35, 0x34, 0xF4, 0x3C, 0xFC, 0xFD, 0x3D, 0xFF, 0x3F, 0x3E, 0xFE, 0xFA, 0x3A,
	0x3B, 0xFB, 0x39, 0xF9, 0xF8, 0x38, 0x28, 0xE8, 0xE9, 0x29, 0xEB, 0x2B, 0x2A, 0xEA, 0xEE,
	0x2E, 0x2F, 0xEF, 0x2D, 0xED, 0xEC, 0x2C, 0xE4, 0x24, 0x25, 0xE5, 0x27, 0xE7, 0xE6, 0x26,
	0x22, 0xE2, 0xE3, 0x23, 0xE1, 0x21, 0x20, 0xE0, 0xA0, 0x60, 0x61, 0xA1, 0x63, 0xA3, 0xA2,
	0x62, 0x66, 0xA6, 0xA7, 0x67, 0xA5, 0x65, 0x64, 0xA4, 0x6C, 0xAC, 0xAD, 0x6D, 0xAF, 0x6F,
	0x6E, 0xAE, 0xAA, 0x6A, 0x6B, 0xAB, 0x69, 0xA9, 0xA8, 0x68, 0x78, 0xB8, 0xB9, 0x79, 0xBB,
	0x7B, 0x7A, 0xBA, 0xBE, 0x7E, 0x7F, 0xBF, 0x7D, 0xBD, 0xBC, 0x7C, 0xB4, 0x74, 0x75, 0xB5,
	0x77, 0xB7, 0xB6, 0x76, 0x72, 0xB2, 0xB3, 0x73, 0xB1, 0x71, 0x70, 0xB0, 0x50, 0x90, 0x91,
	0x51, 0x93, 0x53, 0x52, 0x92, 0x96, 0x56, 0x57, 0x97, 0x55, 0x95, 0x94, 0x54, 0x9C, 0x5C,
	0x5D, 0x9D, 0x5F, 0x9F, 0x9E, 0x5E, 0x5A, 0x9A, 0x9B, 0x5B, 0x99, 0x59, 0x58, 0x98, 0x88,
	0x48, 0x49, 0x89, 0x4B, 0x8B, 0x8A, 0x4A, 0x4E, 0x8E, 0x8F, 0x4F, 0x8D, 0x4D, 0x4C, 0x8C,
	0x44, 0x84, 0x85, 0x45, 0x87, 0x47, 0x46, 0x86, 0x82, 0x42, 0x43, 0x83, 0x41, 0x81, 0x80,
	0x40
} ;


#define T35  5 

/**
* @struct modbus_t
* @brief
* Master query structure:
* This includes all the necessary fields to make the Master generate a Modbus query.
* A Master may keep several of these structures and send them cyclically or
* use them according to program needs.
*/
typedef struct
{
	uint8_t u8id;          /*!< Slave address between 1 and 247. 0 means broadcast */
	uint8_t u8fct;         /*!< Function code: 1, 2, 3, 4, 5, 6, 15 or 16 */
	uint16_t u16RegAdd;    /*!< Address of the first register to access at slave/s */
	uint16_t u16CoilsNo;   /*!< Number of coils or registers to access */
	uint16_t *au16reg;     /*!< Pointer to memory image in master */
}
modbus_t;

enum
{
	RESPONSE_SIZE = 6,
	EXCEPTION_SIZE = 3,
	CHECKSUM_SIZE = 2
};

/**
* @enum MESSAGE
* @brief
* Indexes to telegram frame positions
*/
enum MESSAGE
{
	ID                             = 0, //!< ID field
	FUNC, //!< Function code position
	ADD_HI, //!< Address high byte
	ADD_LO, //!< Address low byte
	NB_HI, //!< Number of coils or registers high byte
	NB_LO, //!< Number of coils or registers low byte
	BYTE_CNT  //!< byte counter
};

/**
* @enum MB_FC
* @brief
* Modbus function codes summary.
* These are the implement function codes either for Master or for Slave.
*
* @see also fctsupported
* @see also modbus_t
*/
enum MB_FC
{
	MB_FC_NONE                     = 0, /*!< null operator */
	MB_FC_READ_COILS               = 1, /*!< FCT=1 -> read coils or digital outputs */
	MB_FC_READ_DISCRETE_INPUT      = 2, /*!< FCT=2 -> read digital inputs */
	MB_FC_READ_REGISTERS           = 3, /*!< FCT=3 -> read registers or analog outputs */
	MB_FC_READ_INPUT_REGISTER      = 4, /*!< FCT=4 -> read analog inputs */
	MB_FC_WRITE_COIL               = 5, /*!< FCT=5 -> write single coil or output */
	MB_FC_WRITE_REGISTER           = 6, /*!< FCT=6 -> write single register */
	MB_FC_WRITE_MULTIPLE_COILS     = 15,/*!< FCT=15 -> write multiple coils or outputs */
	MB_FC_WRITE_MULTIPLE_REGISTERS = 16 /*!< FCT=16 -> write multiple registers */
};

enum COM_STATES
{
	COM_IDLE                     = 0,
	COM_WAITING                  = 1

};

enum ERR_LIST
{
	ERR_NOT_MASTER                = -1,
	ERR_POLLING                   = -2,
	ERR_BUFF_OVERFLOW             = -3,
	ERR_BAD_CRC                   = -4,
	ERR_EXCEPTION                 = -5
};

enum
{
	NO_REPLY = 255,
	EXC_FUNC_CODE = 1,
	EXC_ADDR_RANGE = 2,
	EXC_REGS_QUANT = 3,
	EXC_EXECUTE = 4
};

const uint8_t fctsupported[] =
{
	MB_FC_READ_COILS,
	MB_FC_READ_DISCRETE_INPUT,
	MB_FC_READ_REGISTERS,
	MB_FC_READ_INPUT_REGISTER,
	MB_FC_WRITE_COIL,
	MB_FC_WRITE_REGISTER,
	MB_FC_WRITE_MULTIPLE_COILS,
	MB_FC_WRITE_MULTIPLE_REGISTERS
};

/* _____PUBLIC FUNCTIONS_____________________________________________________ */  
/**
* @brief
* Full constructor for a Master/Slave through USB/RS232C/RS485
* It needs a pin for flow control only for RS485 mode
*
* @param u8id   node address 0=master, 1..247=slave
* @param u8serno  serial port used 0..3
* @param u8txenpin pin for txen RS-485 (=0 means USB/RS232C mode)
* @ingroup setup
* @overload Modbus::Modbus(uint8_t u8id, uint8_t u8serno, uint8_t u8txenpin)
* @overload Modbus::Modbus(uint8_t u8id)
* @overload Modbus::Modbus()
*/
Modbus::Modbus(uint8_t u8id, uint8_t u8serno, uint8_t u8txenpin)
{
	init(u8id, u8serno, u8txenpin);
}

/**
* @brief
* Initialize class object.
*
* Sets up the serial port using specified baud rate.
* Call once class has been instantiated, typically within setup().
*
* @see http://arduino.cc/en/Serial/Begin#.Uy4CJ6aKlHY
* @param speed   baud rate, in standard increments (300..115200)
* @ingroup setup
*/
void Modbus::begin(uint32_t u32speed)
{
	#ifdef USE_AVR_UART
		UARTOpen(u8serno, u32speed);
	#else
	  switch( u8serno )
      {
	  #if defined(UBRR1H)
		  case 1:
			  port = &Serial1;
			  break;
	  #endif
  
	  #if defined(UBRR2H)
		  case 2:
			  port = &Serial2;
			  break;
	  #endif
  
	  #if defined(UBRR3H)
		  case 3:
			  port = &Serial3;
			  break;
	  #endif
		  case 0:
		  default:
			  port = &Serial;
			  break;
	  }
  
	  port->begin(u32speed);
	#endif
	if (u8txenpin > 1)   // pin 0 & pin 1 are reserved for RX/TX
	{
		// return RS485 transceiver to transmit mode
		pinMode(u8txenpin, OUTPUT);
		digitalWrite(u8txenpin, LOW);
	}

	#ifdef USE_AVR_UART
		uint8_t cc = UARTAvailable(u8serno);
		while (cc--) 
		{
			UARTRead(u8serno);
		}
	#else
		while(port->read() >= 0);
	#endif
	u8lastRec = u8BufferSize = 0;
	u16InCnt = u16OutCnt = u16errCnt = 0;
}

/**
* @brief
* Method to write a new slave ID address
*
* @param   u8id  new slave address between 1 and 247
* @ingroup setup
*/
void Modbus::setID(uint8_t u8id)
{
	if (( u8id != 0) && (u8id <= 247))
	{
		this->u8id = u8id;
	}
}

/**
* @brief
* Method to read current slave ID address
*
* @return u8id current slave address between 1 and 247
* @ingroup setup
*/
uint8_t Modbus::getID()
{
	return this->u8id;
}

/**
* @brief
* Initialize time-out parameter
*
* Call once class has been instantiated, typically within setup().
* The time-out timer is reset each time that there is a successful communication
* between Master and Slave. It works for both.
*
* @param time-out value (ms)
* @ingroup setup
*/
void Modbus::setTimeOut( uint16_t u16timeOut)
{
	this->u16timeOut = u16timeOut;
}

/**
* @brief
* Get input messages counter value
* This can be useful to diagnose communication
*
* @return input messages counter
* @ingroup buffer
*/
uint16_t Modbus::getInCnt()
{
	return u16InCnt;
}

/**
* @brief
* Get transmitted messages counter value
* This can be useful to diagnose communication
*
* @return transmitted messages counter
* @ingroup buffer
*/
uint16_t Modbus::getOutCnt()
{
	return u16OutCnt;
}

/**
* @brief
* Get errors counter value
* This can be useful to diagnose communication
*
* @return errors counter
* @ingroup buffer
*/
uint16_t Modbus::getErrCnt()
{
	return u16errCnt;
}

/**
* Get modbus master state
*
* @return = 0 IDLE, = 1 WAITING FOR ANSWER
* @ingroup buffer
*/
uint8_t Modbus::getState()
{
	return u8state;
}

/**
* Get the last error in the protocol processor
*
* @returnreturn   NO_REPLY = 255      Time-out
* @return   EXC_FUNC_CODE = 1   Function code not available
* @return   EXC_ADDR_RANGE = 2  Address beyond available space for Modbus registers
* @return   EXC_REGS_QUANT = 3  Coils or registers number beyond the available space
* @ingroup buffer
*/
uint8_t Modbus::getLastError()
{
	return u8lastError;
}

/**
* @brief
* *** Only for Modbus Slave ***
* This method checks if there is any incoming query
* Afterwards, it would shoot a validation routine plus a register query
* Avoid any delay() function !!!!
* After a successful frame between the Master and the Slave, the time-out timer is reset.
*
* @param *regs  register table for communication exchange
* @param u8size  size of the register table
* @return 0 if no query, 1..4 if communication error, >4 if correct query processed
* @ingroup loop
*/
int8_t Modbus::poll( uint16_t *regs, uint16_t u16SizeArray )
{ 
	if (!stepSlave)
	{
		au16regs = regs;
		u16regsize = u16SizeArray;

		stepSlave = 1;
	}

	if (stepSlave == 1)
	{      
		#ifdef USE_AVR_UART
			uint8_t u8current = UARTAvailable(u8serno);
		#else
			uint8_t u8current;
			if(u8serno<4)
			{
				u8current = port->available();
			}
		#endif
		if (u8current == 0) 
		{
			return 0;
		}

		// check T35 after frame end or still no frame end
		if (u8current != u8lastRec)
		{
			u8lastRec = u8current;
			u32time = millis() + T35;
			return 0;
		}

		if (millis() < u32time)
		{
			return 0;
		}

		u8lastRec = 0;

		stepSlave = 2;
	}

	int8_t i8state = 0;
	if (stepSlave == 2)
	{
		i8state = getRxBuffer();
		u8lastError = i8state;
		if (i8state < 7) 
		{
			return i8state;
		}

		// check slave id
		if (au8Buffer[ ID ] != u8id)
		{
			return 0;
		}

		stepSlave = 3;
	}

	if (stepSlave == 3)
	{
		// validate message: CRC, FCT, address and size
		uint8_t u8exception = validateRequest();
		if (u8exception > 0)
		{
			if (u8exception != NO_REPLY)
			{   
				stepSlave = 4;

				buildException( u8exception );
				sendTxBuffer();
			}
			else 
			{
				stepSlave = 9;
			}

			u8lastError = u8exception;
			return u8exception;
		}
		else 
		{
			stepSlave = 4;

			u8lastError = 0;

			// process message
			switch( au8Buffer[ FUNC ] )
			{
			case MB_FC_READ_COILS:
			case MB_FC_READ_DISCRETE_INPUT:
				return process_FC1( regs, u16SizeArray );
				break;
			case MB_FC_READ_INPUT_REGISTER:
			case MB_FC_READ_REGISTERS :
				return process_FC3( regs, u16SizeArray );
				break;
			case MB_FC_WRITE_COIL:
				return process_FC5( regs, u16SizeArray );
				break;
			case MB_FC_WRITE_REGISTER :
				return process_FC6( regs, u16SizeArray );
				break;
			case MB_FC_WRITE_MULTIPLE_COILS:
				return process_FC15( regs, u16SizeArray );
				break;
			case MB_FC_WRITE_MULTIPLE_REGISTERS :
				return process_FC16( regs, u16SizeArray );
				break;
			default:
				break;
			} 
		}
	}

	if (stepSlave == 5 || stepSlave == 6 || stepSlave == 7)
	{
		sendTxBuffer();
	}

	return i8state;
}

/* _____PRIVATE FUNCTIONS_____________________________________________________ */
void Modbus::init(uint8_t u8id, uint8_t u8serno, uint8_t u8txenpin)
{
	this->u8id = u8id;
	this->u8serno = (u8serno > 3) ? 0 : u8serno;
	this->u8txenpin = u8txenpin;
	this->u16timeOut = 1000;
}

void Modbus::init(uint8_t u8id)
{
	this->u8id = u8id;
	this->u8serno = 4;
	this->u8txenpin = 0;
	this->u16timeOut = 1000;
}

/**
* @brief
* This method moves Serial buffer data to the Modbus au8Buffer.
*
* @return buffer size if OK, ERR_BUFF_OVERFLOW if u8BufferSize >= MAX_BUFFER
* @ingroup buffer
*/
int8_t Modbus::getRxBuffer()
{
	uint8_t bBuffOverflow = 0;

	if (u8txenpin > 1)
	{
		digitalWrite( u8txenpin, LOW );
	}

	u8BufferSize = 0;

	#ifdef USE_AVR_UART
		uint8_t cc = UARTAvailable(u8serno);
		while (cc--) 
		{
			au8Buffer[ u8BufferSize ] = UARTRead(u8serno);
			u8BufferSize ++;

			if (u8BufferSize >= MAX_BUFFER) 
			{
				bBuffOverflow = true;
			}
		}
	#else
		if(u8serno<4)
		{
          while ( port->available() )
          {
              au8Buffer[ u8BufferSize ] = port->read();
              u8BufferSize ++;
  
              if (u8BufferSize >= MAX_BUFFER) bBuffOverflow = true;
          }
		}
	#endif
	u16InCnt++;

	if (bBuffOverflow)
	{
		u16errCnt++;
		return ERR_BUFF_OVERFLOW;
	}
	return u8BufferSize;
}

/**
* @brief
* This method transmits au8Buffer to Serial line.
* Only if u8txenpin != 0, there is a flow handling in order to keep
* the RS485 transceiver in output state as long as the message is being sent.
* This is done with UCSRxA register.
* The CRC is appended to the buffer before starting to send it.
*
* @param nothing
* @return nothing
* @ingroup buffer
*/
void Modbus::sendTxBuffer()
{    
	uint8_t i = 0;
	if (stepSlave == 4)
	{
		// append CRC to message
		uint16_t u16crc = CRC16(au8Buffer, u8BufferSize);

		au8Buffer[ u8BufferSize ] = u16crc >> 8;
		u8BufferSize++;
		au8Buffer[ u8BufferSize ] = u16crc & 0x00ff;
		u8BufferSize++;

		// set RS485 transceiver to transmit mode
		if (u8txenpin > 1)
		{
			#ifndef USE_AVR_UART
				switch( u8serno )
				{
				  #if defined(UBRR1H)
						  case 1:
							  UCSR1A=UCSR1A |(1 << TXC1);
							  break;
				  #endif
  
				  #if defined(UBRR2H)
						  case 2:
							  UCSR2A=UCSR2A |(1 << TXC2);
							  break;
				  #endif
  
				  #if defined(UBRR3H)
						  case 3:
							  UCSR3A=UCSR3A |(1 << TXC3);
							  break;
				  #endif
						  case 0:
						  default:
							  UCSR0A=UCSR0A |(1 << TXC0);
							  break;
				}
			#endif
			digitalWrite( u8txenpin, HIGH );
			u32TimeR485ChangeReDe = micros() + 50UL;
		}
		stepSlave = 5;
	}

	if (stepSlave == 5)
	{ 
		if (u32TimeR485ChangeReDe <= micros())
		{
			// transfer buffer to serial line
			#ifdef USE_AVR_UART
				SerializeStruct(au8Buffer, u8BufferSize);
				UARTSendData(u8serno);       
			#else
				if(u8serno<4)
					port->write( au8Buffer, u8BufferSize );
			#endif

			u32TimeR485ChangeReDe = 0;

			stepSlave = 6;
		}
	}

	if (stepSlave == 6)
	{
		// keep RS485 transceiver in transmit mode as long as sending
		if (u8txenpin > 1)
		{
			#ifdef USE_AVR_UART
				if (!IsUARTFinishSend(u8serno))
				{
					return;
				}
			#else
				switch( u8serno )
				{
				  #if defined(UBRR1H)
						  case 1:
							  if (!(UCSR1A & (1 << TXC1)))
							  {
								return;
							  }
							  break;
				  #endif
  
				  #if defined(UBRR2H)
						  case 2:
							  if (!(UCSR2A & (1 << TXC2)))
							  {
								return;
							  }
							  break;
				  #endif
  
				  #if defined(UBRR3H)
						  case 3:
							  if (!(UCSR3A & (1 << TXC3)))
							  {
								return;
							  }
							  break;
				  #endif
						  case 0:
						  default:
							  if (!(UCSR0A & (1 << TXC0)))
							  {
								return;
							  }
							  break;
				}
			#endif

			u32TimeR485ChangeReDe = micros() + 50UL;
		}
		stepSlave = 7;
	}

	if (stepSlave == 7)
	{
		// return RS485 transceiver to receive mode
		if (u32TimeR485ChangeReDe <= micros())
		{
			u32TimeR485ChangeReDe = 0;

			stepSlave = 8;
		}
	}

	if (stepSlave == 8)
	{
		digitalWrite( u8txenpin, LOW );

		stepSlave = 9;
	}

	if (stepSlave == 9)
	{
		#ifdef USE_AVR_UART
			uint8_t availableByte = UARTAvailable(u8serno);
			while (availableByte--)
			{
				UARTRead(u8serno);
			}
		#else
			if(u8serno<4)
				while(port->read() >= 0);
		#endif

		u8BufferSize = 0;

		// increase message counter
		u16OutCnt++;

		stepSlave = 0;  
	}
}

/**
* @brief
* This method calculates CRC
*
* @return uint16_t calculated CRC value for the message
* @ingroup buffer
*/
uint16_t Modbus::calcCRC(uint8_t u8length)
{
	uint16_t temp, temp2, flag;
	temp = 0xFFFF;
	for (uint8_t i = 0; i < u8length; i++)
	{
		temp = temp ^ au8Buffer[i];
		for (uint8_t j = 1; j <= 8; j++)
		{
			flag = temp & 0x0001;
			temp >>=1;
			if (flag)
				temp ^= 0xA001;
		}
	}
	// Reverse byte order.
	temp2 = temp >> 8;
	temp = (temp << 8) | temp2;
	temp &= 0xFFFF;
	// the returned value is already swapped
	// crcLo byte is first & crcHi byte is last
	return temp;
}

uint16_t Modbus::CRC16(uint8_t *puchMsg, uint16_t usDataLen)
{
	unsigned char uchCRCHi = 0xFF ; /* high byte of CRC initialized */
	unsigned char uchCRCLo = 0xFF ; /* low byte of CRC initialized */
	unsigned uIndex ; /* will index into CRC lookup table */
	while (usDataLen--) /* pass through message buffer */
	{
		uIndex = uchCRCHi ^ *puchMsg++ ; /* calculate the CRC */
		uchCRCHi = uchCRCLo ^ (uint8_t)pgm_read_byte(&auchCRCHi[uIndex]);
		uchCRCLo = (uint8_t)pgm_read_byte(&auchCRCLo[uIndex]);
	}
	return (uchCRCHi << 8 | uchCRCLo);
}

/**
* @brief
* This method validates slave incoming messages
*
* @return 0 if OK, EXCEPTION if anything fails
* @ingroup buffer
*/
uint8_t Modbus::validateRequest()
{
	// check message crc vs calculated crc
	uint16_t u16MsgCrc =
		((au8Buffer[u8BufferSize - 2] << 8)
		| au8Buffer[u8BufferSize - 1]); // combine the crc Low & High bytes


	uint16_t u16MsgCrcCheck = CRC16(au8Buffer, u8BufferSize - 2);
	if (u16MsgCrcCheck != u16MsgCrc )
	{
		u16errCnt ++;
		return NO_REPLY;
	}

	// check fct code
	uint8_t isSupported = 0;
	for (uint8_t i = 0; i< sizeof( fctsupported ); i++)
	{
		if (fctsupported[i] == au8Buffer[FUNC])
		{
			isSupported = 1;
			break;
		}
	}
	if (!isSupported)
	{
		u16errCnt ++;
		return EXC_FUNC_CODE;
	}

	// check start address & nb range
	uint16_t u16regs = 0;
	uint16_t u16CheckRegs;
	switch ( au8Buffer[ FUNC ] )
	{
	case MB_FC_READ_COILS:
	case MB_FC_READ_DISCRETE_INPUT:
	case MB_FC_WRITE_MULTIPLE_COILS:
		u16regs = word( au8Buffer[ ADD_HI ], au8Buffer[ ADD_LO ]) / 16;
		u16regs += word( au8Buffer[ NB_HI ], au8Buffer[ NB_LO ]) /16;
		u16CheckRegs = u16regs;
		if (u16CheckRegs > u16regsize) return EXC_ADDR_RANGE;
		break;
	case MB_FC_WRITE_COIL:
		u16regs = word( au8Buffer[ ADD_HI ], au8Buffer[ ADD_LO ]) / 16;
		u16CheckRegs = u16regs;
		if (u16CheckRegs > u16regsize) return EXC_ADDR_RANGE;
		break;
	case MB_FC_WRITE_REGISTER :
		u16regs = word( au8Buffer[ ADD_HI ], au8Buffer[ ADD_LO ]);
		u16CheckRegs = u16regs;
		if (u16CheckRegs > u16regsize) return EXC_ADDR_RANGE;
		break;
	case MB_FC_READ_REGISTERS :
	case MB_FC_READ_INPUT_REGISTER :
	case MB_FC_WRITE_MULTIPLE_REGISTERS :
		u16regs = word( au8Buffer[ ADD_HI ], au8Buffer[ ADD_LO ]);
		u16regs += word( au8Buffer[ NB_HI ], au8Buffer[ NB_LO ]);
		u16CheckRegs = u16regs;
		if (u16CheckRegs > u16regsize) return EXC_ADDR_RANGE;
		break;
	}
	return 0; // OK, no exception code thrown
}

/**
* @brief
* This method validates master incoming messages
*
* @return 0 if OK, EXCEPTION if anything fails
* @ingroup buffer
*/
uint8_t Modbus::validateAnswer()
{
	// check message crc vs calculated crc
	uint16_t u16MsgCRC =
		((au8Buffer[u8BufferSize - 2] << 8)
		| au8Buffer[u8BufferSize - 1]); // combine the crc Low & High bytes


	uint16_t u16MsgCrcCheck = CRC16(au8Buffer, u8BufferSize-2);
	if (u16MsgCrcCheck != u16MsgCRC )
	{
		u16errCnt ++;
		return NO_REPLY;
	}

	// check exception
	if ((au8Buffer[ FUNC ] & 0x80) != 0)
	{
		u16errCnt ++;
		return ERR_EXCEPTION;
	}

	// check fct code
	uint8_t isSupported = 0;
	for (uint8_t i = 0; i< sizeof( fctsupported ); i++)
	{
		if (fctsupported[i] == au8Buffer[FUNC])
		{
			isSupported = 1;
			break;
		}
	}
	if (!isSupported)
	{
		u16errCnt ++;
		return EXC_FUNC_CODE;
	}

	return 0; // OK, no exception code thrown
}

/**
* @brief
* This method builds an exception message
*
* @ingroup buffer
*/
void Modbus::buildException( uint8_t u8exception )
{
	uint8_t u8func = au8Buffer[ FUNC ];  // get the original FUNC code

	au8Buffer[ ID ]      = u8id;
	au8Buffer[ FUNC ]    = u8func + 0x80;
	au8Buffer[ 2 ]       = u8exception;
	u8BufferSize         = EXCEPTION_SIZE;
}

/**
* @brief
* This method processes functions 1 & 2
* This method reads a bit array and transfers it to the master
*
* @return u8BufferSize Response to master length
* @ingroup discrete
*/
int8_t Modbus::process_FC1( uint16_t *regs, uint16_t u16SizeArray )
{
	uint16_t u16currentRegister, u16bytesno;
	uint8_t u8currentBit, u8bitsno;
	uint8_t u8CopyBufferSize;
	uint16_t u16currentCoil, u16coil;

	// get the first and last coil from the message
	uint16_t u16StartCoil = word( au8Buffer[ ADD_HI ], au8Buffer[ ADD_LO ] );
	uint16_t u16Coilno = word( au8Buffer[ NB_HI ], au8Buffer[ NB_LO ] );

	// put the number of bytes in the outcoming message
	u16bytesno = (u16Coilno / 8);
	if (u16Coilno % 8 != 0) u16bytesno ++;
	au8Buffer[ ADD_HI ]  = highByte(u16bytesno);
	au8Buffer[ ADD_LO ]  = lowByte(u16bytesno);
	u8BufferSize         = 3;

	// read each coil from the register map and put its value inside the outcoming message
	u8bitsno = 0;

	for (u16currentCoil = 0; u16currentCoil < u16Coilno; u16currentCoil++)
	{
		u16coil = u16StartCoil + u16currentCoil;
		u16currentRegister = (u16coil / 16);
		u8currentBit = (uint8_t) (u16coil % 16);

		bitWrite(
			au8Buffer[ u8BufferSize ],
			u8bitsno,
			bitRead( regs[ u16currentRegister ], u8currentBit ) );
		u8bitsno ++;

		if (u8bitsno > 7)
		{
			u8bitsno = 0;
			u8BufferSize++;
		}
	}

	// send outcoming message
	if (u16Coilno % 8 != 0) u8BufferSize ++;
	u8CopyBufferSize = u8BufferSize +2;
	sendTxBuffer();
	return u8CopyBufferSize;
}

/**
* @brief
* This method processes functions 3 & 4
* This method reads a word array and transfers it to the master
*
* @return u8BufferSize Response to master length
* @ingroup register
*/
int8_t Modbus::process_FC3( uint16_t *regs, uint16_t u16SizeArray )
{

	uint16_t u16StartAdd = word( au8Buffer[ ADD_HI ], au8Buffer[ ADD_LO ] );
	uint16_t u16regsno = word( au8Buffer[ NB_HI ], au8Buffer[ NB_LO ] );
	uint8_t u8CopyBufferSize;
	uint8_t i;

	au8Buffer[ 2 ]       = u16regsno;
	u8BufferSize         = 3;

	for (i = u16StartAdd; i < u16StartAdd + u16regsno; i++)
	{
		au8Buffer[ u8BufferSize ] = highByte(regs[i]);
		u8BufferSize++;
		au8Buffer[ u8BufferSize ] = lowByte(regs[i]);
		u8BufferSize++;
	}
	u8CopyBufferSize = u8BufferSize +2;
	sendTxBuffer();

	return u8CopyBufferSize;
}

/**
* @brief
* This method processes function 5
* This method writes a value assigned by the master to a single bit
*
* @return u8BufferSize Response to master length
* @ingroup discrete
*/
int8_t Modbus::process_FC5( uint16_t *regs, uint16_t u16SizeArray )
{
	uint8_t u16currentRegister;
	uint8_t u8currentBit;
	uint8_t u8CopyBufferSize;
	uint16_t u16coil = word( au8Buffer[ ADD_HI ], au8Buffer[ ADD_LO ] );

	// point to the register and its bit
	u16currentRegister = (u16coil / 16);
	u8currentBit = (uint8_t) (u16coil % 16);

	// write to coil
	if (bitRead(regs[ u16currentRegister ], u8currentBit) != (au8Buffer[ NB_HI ] == 0xff))
	{
		bitWrite(
			regs[ u16currentRegister ],
			u8currentBit,
			au8Buffer[ NB_HI ] == 0xff );

		#ifdef MODBUS_SAVE_TO_EEPROM
			uint16_t addressEeprom = (uint16_t)u16currentRegister*(uint16_t)2;
			EepromWriteWord((uint16_t*)addressEeprom, regs[ u16currentRegister ]);
		#endif
	} 

	// send answer to master
	u8BufferSize = 6;
	u8CopyBufferSize = u8BufferSize +2;
	sendTxBuffer();

	return u8CopyBufferSize;
}

/**
* @brief
* This method processes function 6
* This method writes a value assigned by the master to a single word
*
* @return u8BufferSize Response to master length
* @ingroup register
*/
int8_t Modbus::process_FC6( uint16_t *regs, uint16_t u16SizeArray )
{

	uint16_t u16add = word( au8Buffer[ ADD_HI ], au8Buffer[ ADD_LO ] );
	uint8_t u8CopyBufferSize;
	uint16_t u16val = word( au8Buffer[ NB_HI ], au8Buffer[ NB_LO ] );

	// Сохранение в память.
	if (regs[u16add] != u16val)
	{
		regs[ u16add ] = u16val;

		#ifdef MODBUS_SAVE_TO_EEPROM
			uint16_t addressEeprom = (uint16_t)u16add*(uint16_t)2;
			EepromWriteWord((uint16_t*)addressEeprom, u16val);
		#endif
	}

	// keep the same header
	u8BufferSize         = RESPONSE_SIZE;

	u8CopyBufferSize = u8BufferSize +2;
	sendTxBuffer();

	return u8CopyBufferSize;
}

/**
* @brief
* This method processes function 15
* This method writes a bit array assigned by the master
*
* @return u8BufferSize Response to master length
* @ingroup discrete
*/
int8_t Modbus::process_FC15( uint16_t *regs, uint16_t u16SizeArray )
{
	uint8_t u16currentRegister;
	uint8_t u8currentBit, u8frameByte, u8bitsno;
	uint8_t u8CopyBufferSize;
	uint16_t u16currentCoil, u16coil;
	boolean bTemp;

	// get the first and last coil from the message
	uint16_t u16StartCoil = word( au8Buffer[ ADD_HI ], au8Buffer[ ADD_LO ] );
	uint16_t u16Coilno = word( au8Buffer[ NB_HI ], au8Buffer[ NB_LO ] );


	// read each coil from the register map and put its value inside the outcoming message
	u8bitsno = 0;
	u8frameByte = 7;
	for (u16currentCoil = 0; u16currentCoil < u16Coilno; u16currentCoil++)
	{

		u16coil = u16StartCoil + u16currentCoil;
		u16currentRegister = (u16coil / 16);
		u8currentBit = (uint8_t) (u16coil % 16);

		bTemp = bitRead(
			au8Buffer[ u8frameByte ],
			u8bitsno );

		// Сохранение в память.
		if (bitRead(regs[ u16currentRegister ], u8currentBit) != bTemp)
		{
			bitWrite(
				regs[ u16currentRegister ],
				u8currentBit,
				bTemp );

			#ifdef MODBUS_SAVE_TO_EEPROM
				uint16_t addressEeprom = (uint16_t)u16currentRegister*(uint16_t)2;
				EepromWriteWord((uint16_t*)addressEeprom, regs[ u16currentRegister ]);
			#endif
		}

		u8bitsno ++;

		if (u8bitsno > 7)
		{
			u8bitsno = 0;
			u8frameByte++;
		}
	}

	// send outcoming message
	// it's just a copy of the incomping frame until 6th byte
	u8BufferSize         = 6;
	u8CopyBufferSize = u8BufferSize +2;
	sendTxBuffer();
	return u8CopyBufferSize;
}

/**
* @brief
* This method processes function 16
* This method writes a word array assigned by the master
*
* @return u8BufferSize Response to master length
* @ingroup register
*/
int8_t Modbus::process_FC16( uint16_t *regs, uint16_t u16SizeArray )
{
	uint8_t u8func = au8Buffer[ FUNC ];  // get the original FUNC code
	uint16_t u16StartAdd = au8Buffer[ ADD_HI ] << 8 | au8Buffer[ ADD_LO ];
	uint16_t u16regsno = au8Buffer[ NB_HI ] << 8 | au8Buffer[ NB_LO ];
	uint8_t u8CopyBufferSize;
	uint8_t i;
	uint16_t temp;

	// build header
	au8Buffer[ NB_HI ]   = highByte(u16regsno);
	au8Buffer[ NB_LO ]   = lowByte(u16regsno);
	u8BufferSize         = RESPONSE_SIZE;

	// write registers
	for (i = 0; i < u16regsno; i++)
	{
		temp = word(
			au8Buffer[ (BYTE_CNT + 1) + i * 2 ],
			au8Buffer[ (BYTE_CNT + 2) + i * 2 ]);

		if (regs[ u16StartAdd + i ] != temp)
		{
			regs[ u16StartAdd + i ] = temp;

			#ifdef MODBUS_SAVE_TO_EEPROM
				uint16_t addressEeprom = (uint16_t)(u16StartAdd + i)*(uint16_t)2;
				EepromWriteWord((uint16_t*)addressEeprom, temp);
			#endif
		}  
	}
	u8CopyBufferSize = u8BufferSize +2;
	sendTxBuffer();

	return u8CopyBufferSize;
}

#ifdef USE_AVR_UART
	static void Modbus::SerializeStruct(uint8_t *cb, uint8_t siz) 
	{
		while(siz--) 
		{
			UARTSerialize(MODBUS_UART_PORT, *cb++);
		}
	}
#endif
