/**
 * @file   ModbusRtu.h
 * @version     1.21
 * @date        2016.02.21
 * @author  Samuel Marco i Armengol
 * @contact     sammarcoarmengol@gmail.com
 * @contribution Helium6072
 *
 * @description
 *  Arduino library for communicating with Modbus devices
 *  over RS232/USB/485 via RTU protocol.
 *
 *  Further information:
 *  http://modbus.org/
 *  http://modbus.org/docs/Modbus_over_serial_line_V1_02.pdf
 *
 * @license
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; version
 *  2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * @defgroup setup Modbus Object Instantiation/Initialization
 * @defgroup loop Modbus Object Management
 * @defgroup buffer Modbus Buffer Management
 * @defgroup discrete Modbus Function Codes for Discrete Coils/Inputs
 * @defgroup register Modbus Function Codes for Holding/Input Registers
 *
 */

/******************************************************************************
     * Based on the Modbus Master-Slave library for Arduino. 
     * Author project library - smarmengol.
     * Version on commit 8321f2dc15535d8cecd4a08331aaf112f171c31c on 30 Jul 2016.
     * Modbus Master-Slave library for Arduino - https://github.com/smarmengol/Modbus-Master-Slave-for-Arduino. 
     * GNU Lesser General Public License v2.1
******************************************************************************/

/******************************************************************************
     * File: ModbusRtuSlave.h
     * Description: Библиотека работы с Modbus RTU в режиме Slave.
     * Created: 22 ноября 2016
     * Last Modifiede: 06 декабря 2016 
     * Copyright: GNU General Public License
******************************************************************************/

/*
//TODO: Добавить комментарии.
//TODO: Сделать буфер для отложенной записи данных и реализацию для отложенной записи в EEPROM.
*/

//*******************************  Библиотеки  ******************************//
#include "Arduino.h"

#include "Config.h"
//*******************************  /Библиотеки  *****************************//
#define  MAX_BUFFER  64 //!< maximum size for the communication buffer in bytes

/**
* @class Modbus
* @brief
* Arduino class library for communicating with Modbus devices over
* USB/RS232/485 (via RTU protocol).
*/
class Modbus
{
private:
	#ifndef USE_AVR_UART
		HardwareSerial *port; //!< Pointer to Serial class object
	#endif
	uint8_t u8id; //!< 0=master, 1..247=slave number
	uint8_t u8serno; //!< serial port: 0-Serial, 1..3-Serial1..Serial3; 4: use software serial
	uint8_t u8txenpin; //!< flow control pin: 0=USB or RS-232 mode, >0=RS-485 mode
	uint8_t u8state;
	uint8_t u8lastError;
	uint8_t au8Buffer[MAX_BUFFER];
	uint8_t u8BufferSize;
	uint8_t u8lastRec;
	uint16_t *au16regs;
	uint16_t u16InCnt, u16OutCnt, u16errCnt;
	uint16_t u16timeOut;
	uint32_t u32time, u32TimeR485ChangeReDe;
	uint16_t u16regsize;

	void init(uint8_t u8id, uint8_t u8serno, uint8_t u8txenpin);
	#ifdef USE_AVR_UART
		static void Modbus::SerializeStruct(uint8_t *cb,uint8_t siz);
	#endif
	uint8_t stepSlave; 
	void init(uint8_t u8id);
	void sendTxBuffer();
	int8_t getRxBuffer();
	uint16_t calcCRC(uint8_t u8length);
	uint16_t CRC16(uint8_t *puchMsg, uint16_t usDataLen);
	uint8_t validateAnswer();
	uint8_t validateRequest();
	int8_t process_FC1( uint16_t *regs, uint16_t u16SizeArray );
	int8_t process_FC3( uint16_t *regs, uint16_t u16SizeArray );
	int8_t process_FC5( uint16_t *regs, uint16_t u16SizeArray );
	int8_t process_FC6( uint16_t *regs, uint16_t u16SizeArray );
	int8_t process_FC15( uint16_t *regs, uint16_t u16SizeArray );
	int8_t process_FC16( uint16_t *regs, uint16_t u16SizeArray );
	void buildException( uint8_t u8exception ); // build exception message

public:
	Modbus(uint8_t u8id, uint8_t u8serno, uint8_t u8txenpin);
	void begin(uint32_t u32speed);
	void setTimeOut( uint16_t u16timeout); //!<write communication watch-dog timer
	uint16_t getTimeOut(); //!<get communication watch-dog timer value
	int8_t poll( uint16_t *regs, uint16_t u16SizeArray ); //!<cyclic poll for slave
	uint16_t getInCnt(); //!<number of incoming messages
	uint16_t getOutCnt(); //!<number of outcoming messages
	uint16_t getErrCnt(); //!<error counter
	uint8_t getID(); //!<get slave ID between 1 and 247
	uint8_t getState();
	uint8_t getLastError(); //!<get last error message
	void setID( uint8_t u8id ); //!<write new ID for the slave
	void end(); //!<finish any communication and release serial communication port
};

